@@ -23,26 +23,49 @@
 #include "iw.h"
 #include "version.h"
 
-int debug = 0;
+#ifndef CONFIG_LIBNL20
+/* libnl 2.0 compatibility code */
+
+static inline struct nl_handle *nl_socket_alloc(void)
+{
+	return nl_handle_alloc();
+}
+
+static inline void nl_socket_free(struct nl_sock *h)
+{
+	nl_handle_destroy(h);
+}
+
+static inline int __genl_ctrl_alloc_cache(struct nl_sock *h, struct nl_cache **cache)
+{
+	struct nl_cache *tmp = genl_ctrl_alloc_cache(h);
+	if (!tmp)
+		return -ENOMEM;
+	*cache = tmp;
+	return 0;
+}
+#define genl_ctrl_alloc_cache __genl_ctrl_alloc_cache
+#endif /* CONFIG_LIBNL20 */
+
+static int debug = 0;
 
 static int nl80211_init(struct nl80211_state *state)
 {
 	int err;
 
-	state->nl_handle = nl_handle_alloc();
-	if (!state->nl_handle) {
-		fprintf(stderr, "Failed to allocate netlink handle.\n");
+	state->nl_sock = nl_socket_alloc();
+	if (!state->nl_sock) {
+		fprintf(stderr, "Failed to allocate netlink socket.\n");
 		return -ENOMEM;
 	}
 
-	if (genl_connect(state->nl_handle)) {
+	if (genl_connect(state->nl_sock)) {
 		fprintf(stderr, "Failed to connect to generic netlink.\n");
 		err = -ENOLINK;
 		goto out_handle_destroy;
 	}
 
-	state->nl_cache = genl_ctrl_alloc_cache(state->nl_handle);
-	if (!state->nl_cache) {
+	if (genl_ctrl_alloc_cache(state->nl_sock, &state->nl_cache)) {
 		fprintf(stderr, "Failed to allocate generic netlink cache.\n");
 		err = -ENOMEM;
 		goto out_handle_destroy;
@@ -60,7 +83,7 @@ static int nl80211_init(struct nl80211_state *state)
  out_cache_free:
 	nl_cache_free(state->nl_cache);
  out_handle_destroy:
-	nl_handle_destroy(state->nl_handle);
+	nl_socket_free(state->nl_sock);
 	return err;
 }
 
@@ -68,7 +91,7 @@ static void nl80211_cleanup(struct nl80211_state *state)
 {
 	genl_family_put(state->nl80211);
 	nl_cache_free(state->nl_cache);
-	nl_handle_destroy(state->nl_handle);
+	nl_socket_free(state->nl_sock);
 }
 
 __COMMAND(NULL, NULL, NULL, 0, 0, 0, CIB_NONE, NULL);
@@ -126,6 +149,8 @@ static int phy_lookup(char *name)
 	snprintf(buf, sizeof(buf), "/sys/class/ieee80211/%s/index", name);
 
 	fd = open(buf, O_RDONLY);
+	if (fd < 0)
+		return -1;
 	pos = read(fd, buf, sizeof(buf) - 1);
 	if (pos < 0)
 		return -1;
@@ -155,8 +180,15 @@ static int ack_handler(struct nl_msg *msg, void *arg)
 	return NL_STOP;
 }
 
+enum id_input {
+	II_NONE,
+	II_NETDEV,
+	II_PHY_NAME,
+	II_PHY_IDX,
+};
+
 static int handle_cmd(struct nl80211_state *state,
-		      enum command_identify_by idby,
+		      enum id_input idby,
 		      int argc, char **argv)
 {
 	struct cmd *cmd;
@@ -165,18 +197,32 @@ static int handle_cmd(struct nl80211_state *state,
 	int devidx = 0;
 	int err;
 	const char *command, *section;
+	char *tmp;
+	enum command_identify_by command_idby = CIB_NONE;
 
-	if (argc <= 1 && idby != CIB_NONE)
+	if (argc <= 1 && idby != II_NONE)
 		return 1;
 
 	switch (idby) {
-	case CIB_PHY:
+	case II_PHY_IDX:
+		command_idby = CIB_PHY;
+		devidx = strtoul(*argv + 4, &tmp, 0);
+		if (*tmp != '\0')
+			return 1;
+		argc--;
+		argv++;
+		break;
+	case II_PHY_NAME:
+		command_idby = CIB_PHY;
 		devidx = phy_lookup(*argv);
 		argc--;
 		argv++;
 		break;
-	case CIB_NETDEV:
+	case II_NETDEV:
+		command_idby = CIB_NETDEV;
 		devidx = if_nametoindex(*argv);
+		if (devidx == 0)
+			devidx = -1;
 		argc--;
 		argv++;
 		break;
@@ -184,6 +230,9 @@ static int handle_cmd(struct nl80211_state *state,
 		break;
 	}
 
+	if (devidx < 0)
+		return -errno;
+
 	section = command = *argv;
 	argc--;
 	argv++;
@@ -192,7 +241,7 @@ static int handle_cmd(struct nl80211_state *state,
 	     cmd = (struct cmd *)((char *)cmd + cmd_size)) {
 		if (!cmd->handler)
 			continue;
-		if (cmd->idby != idby)
+		if (cmd->idby != command_idby)
 			continue;
 		if (cmd->section) {
 			if (strcmp(cmd->section, section))
@@ -233,7 +282,7 @@ static int handle_cmd(struct nl80211_state *state,
 	genlmsg_put(msg, 0, 0, genl_family_get_id(state->nl80211), 0,
 		    cmd->nl_msg_flags, cmd->cmd, 0);
 
-	switch (idby) {
+	switch (command_idby) {
 	case CIB_PHY:
 		NLA_PUT_U32(msg, NL80211_ATTR_WIPHY, devidx);
 		break;
@@ -248,7 +297,7 @@ static int handle_cmd(struct nl80211_state *state,
 	if (err)
 		goto out;
 
-	err = nl_send_auto_complete(state->nl_handle, msg);
+	err = nl_send_auto_complete(state->nl_sock, msg);
 	if (err < 0)
 		goto out;
 
@@ -259,7 +308,7 @@ static int handle_cmd(struct nl80211_state *state,
 	nl_cb_set(cb, NL_CB_ACK, NL_CB_CUSTOM, ack_handler, &err);
 
 	while (err > 0)
-		nl_recvmsgs(state->nl_handle, cb);
+		nl_recvmsgs(state->nl_sock, cb);
  out:
 	nl_cb_put(cb);
  out_free_msg:
@@ -279,17 +328,67 @@ static int print_event(struct nl_msg *msg, void *arg)
 {
 	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
 	struct nlattr *tb[NL80211_ATTR_MAX + 1];
+	char ifname[100];
+	__u8 reg_type;
 
 	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
 		  genlmsg_attrlen(gnlh, 0), NULL);
                           
 	switch (gnlh->cmd) {
-	case NL80211_CMD_NEW_WIPHY: {
+	case NL80211_CMD_NEW_WIPHY:
 		printf("wiphy rename: phy #%d to %s\n",
 		       nla_get_u32(tb[NL80211_ATTR_WIPHY]),
 		       nla_get_string(tb[NL80211_ATTR_WIPHY_NAME]));
 		break;
-	}
+	case NL80211_CMD_NEW_SCAN_RESULTS:
+		if_indextoname(nla_get_u32(tb[NL80211_ATTR_IFINDEX]), ifname);
+		printf("scan finished on %s (phy #%d)\n",
+		       ifname, nla_get_u32(tb[NL80211_ATTR_WIPHY]));
+		break;
+	case NL80211_CMD_SCAN_ABORTED:
+		if_indextoname(nla_get_u32(tb[NL80211_ATTR_IFINDEX]), ifname);
+		printf("scan aborted on %s (phy #%d)\n",
+		       ifname, nla_get_u32(tb[NL80211_ATTR_WIPHY]));
+		break;
+	case NL80211_CMD_REG_CHANGE:
+
+		printf("regulatory domain change: ");
+
+		reg_type = nla_get_u8(tb[NL80211_ATTR_REG_TYPE]);
+
+		switch (reg_type) {
+		case NL80211_REGDOM_TYPE_COUNTRY:
+			printf("set to %s by %s request",
+			       nla_get_string(tb[NL80211_ATTR_REG_ALPHA2]),
+			       reg_initiator_to_string(nla_get_u8(tb[NL80211_ATTR_REG_INITIATOR])));
+			if (tb[NL80211_ATTR_WIPHY])
+				printf(" on phy%d", nla_get_u32(tb[NL80211_ATTR_WIPHY]));
+			break;
+		case NL80211_REGDOM_TYPE_WORLD:
+			printf("set to world roaming by %s request",
+			       reg_initiator_to_string(nla_get_u8(tb[NL80211_ATTR_REG_INITIATOR])));
+			break;
+		case NL80211_REGDOM_TYPE_CUSTOM_WORLD:
+			printf("custom world roaming rules in place on phy%d by %s request",
+			       nla_get_u32(tb[NL80211_ATTR_WIPHY]),
+			       reg_initiator_to_string(nla_get_u32(tb[NL80211_ATTR_REG_INITIATOR])));
+			break;
+		case NL80211_REGDOM_TYPE_INTERSECTION:
+			printf("intersection used due to a request made by %s",
+			       reg_initiator_to_string(nla_get_u32(tb[NL80211_ATTR_REG_INITIATOR])));
+			if (tb[NL80211_ATTR_WIPHY])
+				printf(" on phy%d", nla_get_u32(tb[NL80211_ATTR_WIPHY]));
+			break;
+		default:
+			printf("unknown source (upgrade this utility)");
+			break;
+		}
+
+		printf("\n");
+		break;
+	default:
+		printf("unknown event: %d\n", gnlh->cmd);
+		break;
 	}
 
 	return NL_SKIP;
@@ -306,20 +405,37 @@ static int listen_events(struct nl80211_state *state,
 		return -ENOMEM;
 	}
 
-	mcid = nl_get_multicast_id(state->nl_handle, "nl80211", "config");
+	/* Configuration multicast group */
+	mcid = nl_get_multicast_id(state->nl_sock, "nl80211", "config");
 	if (mcid < 0)
 		return mcid;
 
-	ret = nl_socket_add_membership(state->nl_handle, mcid);
+	ret = nl_socket_add_membership(state->nl_sock, mcid);
 	if (ret)
 		return ret;
-	
+
+	/* Scan multicast group */
+	mcid = nl_get_multicast_id(state->nl_sock, "nl80211", "scan");
+	if (mcid >= 0) {
+		ret = nl_socket_add_membership(state->nl_sock, mcid);
+		if (ret)
+			return ret;
+	}
+
+	/* Regulatory multicast group */
+	mcid = nl_get_multicast_id(state->nl_sock, "nl80211", "regulatory");
+	if (mcid >= 0) {
+		ret = nl_socket_add_membership(state->nl_sock, mcid);
+		if (ret)
+			return ret;
+	}
+
 	/* no sequence checking for multicast messages */
 	nl_cb_set(cb, NL_CB_SEQ_CHECK, NL_CB_CUSTOM, no_seq_check, NULL);
 	nl_cb_set(cb, NL_CB_VALID, NL_CB_CUSTOM, print_event, NULL);
 
 	while (1)
-		nl_recvmsgs(state->nl_handle, cb);
+		nl_recvmsgs(state->nl_sock, cb);
 
 	nl_cb_put(cb);
 
@@ -364,13 +480,18 @@ int main(int argc, char **argv)
 	} else if (strcmp(*argv, "dev") == 0) {
 		argc--;
 		argv++;
-		err = handle_cmd(&nlstate, CIB_NETDEV, argc, argv);
-	} else if (strcmp(*argv, "phy") == 0) {
-		argc--;
-		argv++;
-		err = handle_cmd(&nlstate, CIB_PHY, argc, argv);
+		err = handle_cmd(&nlstate, II_NETDEV, argc, argv);
+	} else if (strncmp(*argv, "phy", 3) == 0) {
+		if (strlen(*argv) == 3) {
+			argc--;
+			argv++;
+			err = handle_cmd(&nlstate, II_PHY_NAME, argc, argv);
+		} else if (*(*argv + 3) == '#')
+			err = handle_cmd(&nlstate, II_PHY_IDX, argc, argv);
+		else
+			err = 1;
 	} else
-		err = handle_cmd(&nlstate, CIB_NONE, argc, argv);
+		err = handle_cmd(&nlstate, II_NONE, argc, argv);
 
 	if (err == 1)
 		usage(argv0);
